"""
Fixtures for testing
"""

import pytest
import seamm_datastore

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker


@pytest.fixture()
def session():
    some_engine = create_engine("sqlite:///:memory:")

    # create a configured "Session" class
    Session = sessionmaker(bind=some_engine)

    # create a Session
    sess = Session()
    return sess


@pytest.fixture()
def connection():
    from seamm_datastore import session_scope

    db = seamm_datastore.connect(initialize=True)

    with session_scope(db.Session) as _:
        from seamm_datastore.database.models import User

        users = User.query.all()
        for data in users:
            if data.username != "admin":
                user = data.username
                break
    db.login(username=user, password="default")
    return db


@pytest.fixture()
def admin_connection(connection):
    connection.login(username="admin", password="admin")

    return connection


@pytest.fixture(scope="function")
def filled_db(connection):

    # The lazy way to do this for now
    import os

    loc = os.path.dirname(os.path.abspath(__file__))
    sample_data = os.path.join(loc, "..", "data", "Projects")

    connection.import_datastore(sample_data)



"""
API Tests
"""

import pytest


@pytest.mark.parametrize(
    "resource, description, title, offset, limit, sort, order, num_records, first_id",
    [
        (
            "job",
            None,
            None,
            None,
            None,
            "id",
            "asc",
            2,
            92,
        ),  # Test retrieving all jobs default sort
        (
            "job",
            None,
            None,
            None,
            None,
            "id",
            "desc",
            2,
            93,
        ),  # Test retrieving all jobs, descending sort
        (
            "job",
            None,
            None,
            None,
            None,
            "last_update",
            "desc",
            2,
            93,
        ),  # Test retrieving all jobs, descending sort by last update time.
        (
            "project",
            None,
            None,
            None,
            None,
            "id",
            "asc",
            2,
            1,
        ),  # Test retrieving all projects
        (
            "flowchart",
            None,
            None,
            None,
            None,
            "id",
            "asc",
            2,
            1,
        ),  # Test retrieving all flowcharts
        ("job", None, None, 1, None, "id", "asc", 1, 93),  # Get jobs with offset
        ("job", None, None, None, 1, "id", "asc", 1, 92),  # Get jobs with offset
        ("job", None, "Test", None, None, "id", "asc", 1, 92),
    ],
)
def test_get_queries(
    filled_db,
    resource,
    description,
    title,
    offset,
    limit,
    sort,
    order,
    num_records,
    first_id,
):
    from seamm_datastore.database.models import Job, Project, Flowchart

    resource_map = {"job": Job, "flowchart": Flowchart, "project": Project}

    resource = resource_map[resource]

    records = resource.get(
        description=description,
        title=title,
        offset=offset,
        limit=limit,
        sort_by=sort,
        order=order,
    )

    assert len(records) == num_records

    # Since the test relies on database that is built for testing,
    # we don't know for sure which one is added first. Confirm with
    # times. Not the ideal way :)
    if sort == "last_update":
        if order == "desc":
            assert records[0].last_update >= records[1].last_update
        else:
            assert records[0].last_update <= records[1].last_update
    else:
        assert records[0].id == first_id

    assert isinstance(records[0], resource)


def test_job_update_error(filled_db):

    from seamm_datastore.database.models import Job

    with pytest.raises(ValueError):
        Job.update(1, description="This is a new description.")


def test_job_update(filled_db):

    from seamm_datastore.database.models import Job

    job = Job.update(92, description="This is a new description.")

    assert job.description == "This is a new description."



"""
Test the create functions on the SQLAlchemy models
"""

import os

import pytest


def test_project_create(connection):
    """Test the create method of the project object"""

    project = connection.Project.create(name="test_project")

    assert project.name == "test_project"
    assert project.owner == connection.current_user()

    # If group is not specified, the group should be the
    # first group the user belongs to.
    assert project.group.name == connection.current_user().groups[0].name


def test_project_exists(connection):

    with pytest.raises(ValueError):
        connection.Project.create(name="default")


def test_project_no_user(connection):

    with pytest.raises(ValueError):
        connection.logout()
        connection.Project.create(name="test")


def test_project_create_group(connection):
    project = connection.Project.create(name="test_project", group="admin")

    assert project.name == "test_project"
    assert project.owner == connection.current_user()

    assert project.group.name == "admin"


def test_flowchart_parse(connection):

    this_file = os.path.dirname(os.path.abspath(__file__))
    filepath = os.path.join(this_file, "..", "data", "sample_flowchart_v2.flow")

    metadata, text = connection.Flowchart.parse_flowchart_file(filepath)

    assert metadata["flowchart_version"] == 2.0
    assert (
        metadata["sha256_strict"]
        == "79d580b78559fe137872bcffe24aa7455e6c66fe260cf63e5edd3b3a1464e9c6"
    )

    assert text


def test_flowchart_from_file(connection):

    this_file = os.path.dirname(os.path.abspath(__file__))
    filepath = os.path.join(this_file, "..", "data", "sample_flowchart_v2.flow")

    flowchart = connection.Flowchart.create_from_file(filepath)

    assert (
        flowchart.sha256_strict
        == "79d580b78559fe137872bcffe24aa7455e6c66fe260cf63e5edd3b3a1464e9c6"
    )


def test_create_user(connection):

    user = connection.User.create(username="test", password="test")

    assert user.username == "test"
    assert user.groups[0].name == connection.default_group


def test_add_job(connection):
    from pathlib import Path
    from dateutil import parser

    from seamm_datastore import session_scope

    path = (
        Path(__file__)
        / ".."
        / ".."
        / "data"
        / "Projects"
        / "sample_project1"
        / "Job_000093"
    )
    path = path.expanduser().resolve()

    # connection.add_job(job1_data)
    job = connection.Job.create(
        id=1,
        flowchart_filename=str(path / "flowchart.flow"),
        project_names=["default"],
        path=str(path),
        title="test job",
        description="description of the job",
        submitted=parser.parse("2016-08-29T09:12:33.000000+00:00"),
        started=parser.parse("2016-08-29T09:12:34.000000+00:00"),
        finished=parser.parse("2016-08-29T09:13:34.000000+00:00"),
        status="finished",
    )

    assert job.id == 1

    # Retrieve job
    with session_scope(connection.Session) as sess:
        from seamm_datastore.database.models import Job

        sess.add(job)
        sess.commit()
        jobs = sess.query(Job).all()
        assert len(jobs) == 1



"""
Class and functions for connection to database.
"""

import os

from functools import wraps
from contextlib import contextmanager
from warnings import warn

from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from flask_authorize import Authorize

import seamm_datastore.database.build
from .util import LoginRequiredError


def manage_session(function):
    """Decorator for closing sqlalchemy sessions when attached to SEAMMDatastore
    class.
    """

    def _wrap_method(method):
        def _manage_session(self, *args, **kwargs):
            with session_scope(self.Session) as session:
                ret = function(session, as_json=True, *args, **kwargs)
            return ret

        return _manage_session

    return _wrap_method


@contextmanager
def session_scope(session):
    """Provide a transactional scope around a series of operations."""
    session = session()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()


def login_required(method):
    """Decorator for actions requiring current user"""

    @wraps(method)
    def _check_user(self, *args, **kwargs):
        if not self.current_user():
            raise LoginRequiredError

        else:
            ret = method(self, *args, **kwargs, current_user=self.current_user())
        return ret

    return _check_user


class current_app:
    """Fake current_app"""

    def __init__(self, config):
        self.config = config
        self.extensions = ["sqlalchemy"]


class SEAMMDatastore:
    def __init__(
        self,
        database_uri: str = "sqlite:///:memory:",
        initialize: bool = False,
        permissions: dict = None,
        username: str = None,
        password: str = None,
        datastore_location: str = None,
        default_project: str = "default",
    ):

        if database_uri.lower() == "sqlite:///:memory:":
            initialize = True

        # Default permissions
        if permissions is None:
            permissions = {
                "owner": ["read", "update", "delete"],
                "group": ["read", "update"],
                "world": [],
            }

        global fake_app
        fake_app = current_app(
            config={
                "AUTHORIZE_DEFAULT_PERMISSIONS": permissions,
                "AUTHORIZE_MODEL_PARSER": "table",
                "AUTHORIZE_IGNORE_PROPERTY": "__check_access__",
            }
        )

        # Set datastore location
        if datastore_location is None:
            self.datastore_location = os.getcwd()
        else:
            self.datastore_location = datastore_location

        # Create engine and session
        self.engine = create_engine(database_uri)

        self.Session = scoped_session(
            sessionmaker(autocommit=False, autoflush=False, bind=self.engine)
        )

        from seamm_datastore.database.models import (
            Base,
            Project,
            User,
            Job,
            Group,
            Role,
            Flowchart,
        )

        if initialize:
            Base.metadata.drop_all(self.engine)
        Base.metadata.create_all(self.engine)
        Base.query = self.Session.query_property()

        # Set up the database.
        if initialize:
            seamm_datastore.database.build._build_initial(
                self.Session(), default_project
            )

            group = Group.query.filter_by(name="admin").one()
            admin_role = Role.query.filter_by(name="admin").one()

            # Log in the admin user and create specified user
            if username:
                self.login("admin", "admin")
                self.add_user(username, password, roles=[admin_role], groups=[group])
                self.logout()

        if username:
            self.login(username, password)
        else:
            self._user = None

        # Current user has to be bound before we can add anything else to be database.
        self.authorize = Authorize(current_user=self.current_user)

        # Default group
        self.default_group = Group.query.get(2).name

        # Now handle the project
        project = Project.query.filter_by(name=default_project).one_or_none()
        if not project:
            raise ValueError("Invalid project name given for default.")

        self.default_project = default_project

        self.User = User
        self.Group = Group
        self.Role = Role
        self.Project = Project
        self.Job = Job
        self.Flowchart = Flowchart

    def login(self, username, password):
        from seamm_datastore.database.models import User

        user = User.query.filter_by(username=username).one()

        if not user.verify_password(password):
            raise ValueError("Login unsuccessful. Check username and password.")

        self._user = username

    def logout(self):
        self._user = None

    # Seems like a good place for @property, but can't use because flask authorize
    # requires this to be callable.
    def current_user(self):
        from seamm_datastore.database.models import User

        if self._user:
            user = User.query.filter_by(username=self._user).one_or_none()
        else:
            user = None
        return user

    @manage_session(seamm_datastore.database.build.import_datastore)
    def import_datastore(self, *args, **kwargs):
        pass

    def add_job(
        self,
        id,
        flowchart_filename,
        project_names=["default"],
        path=None,
        title="",
        description="",
        submitted=None,
        started=None,
        finished=None,
        status="submitted",
    ):

        warn(
            "Deprecation warning: This method will no longer be available \
            in the next version of the seamm datastore. \
            Job.create should be used instead."
        )

        with session_scope(self.Session) as session:

            job = self.Job.create(
                id,
                flowchart_filename,
                project_names,
                path,
                title,
                description,
                submitted,
                started,
                finished,
                status,
            )

            session.add(job)

    def finish_job(
        self,
        job_id,
        finish_time,
        status="finished",
    ):
        """Set the status and time that the job finished.

        Parameters
        ----------
        job_id : int
            The ID of the job, eg. 209
        finish_time : datetime.datetime
            The UTC time when the job finished.
        status : str
            The status, such as "error" or the default, "finished"
        as_json : bool = False
            Ignored
        current_user : str or User = None
            Ignored

        Returns
        -------
        bool
            True if the finish time was successfully set, False otherwise.
        """
        warn(
            "Deprecation warning: This method will no longer be available \
            in the next version of the seamm datastore. \
            Job.update should be used instead."
        )

        from seamm_datastore.database.models import Job

        with session_scope(self.Session) as session:
            job = Job.get_by_id(job_id, permission="update")

            if job is None:
                return False
            else:
                job.finished = finish_time
                job.status = status
                session.commit()
                return True
